from modeltranslation.translator import translator, TranslationOptions
from .models import OpenIDClient, OpenIDAgreement


class OpenIDClientTranslationOptions(TranslationOptions):
    fields = ('client_name',)


class OpenIDAgreementTranslationOptions(TranslationOptions):
    fields = ('text',)


translator.register(OpenIDClient, OpenIDClientTranslationOptions)
translator.register(OpenIDAgreement, OpenIDAgreementTranslationOptions)



from openid_connect_op.models import OpenIDClient


class TestRedirectURIMatching:
    def test_exact_match(self):
        client_config = OpenIDClient(redirect_uris='http://my-site.com/auth/complete')
        assert client_config.check_redirect_url('http://my-site.com/auth/complete')

    def test_exact_match_two_uris(self):
        client_config = OpenIDClient(redirect_uris='http://my-site.com/auth/complete\n  http://my-site1.com/auth/complete')
        assert client_config.check_redirect_url('http://my-site.com/auth/complete')
        assert client_config.check_redirect_url('http://my-site1.com/auth/complete')

    def test_non_match(self):
        client_config = OpenIDClient(redirect_uris='http://my-site.com/auth/complete')
        assert not client_config.check_redirect_url('http://my-site.org/auth/complete')
        assert not client_config.check_redirect_url('http://my-site.com:80/auth/complete')
        assert not client_config.check_redirect_url('http://my-site.com:8000/auth/complete')

    def test_match_with_param_with_value(self):
        client_config = OpenIDClient(redirect_uris='http://my-site.com/auth/complete?a=2')
        assert client_config.check_redirect_url('http://my-site.com/auth/complete?a=2')
        assert not client_config.check_redirect_url('http://my-site.com/auth/complete?a=3')
        assert not client_config.check_redirect_url('http://my-site.com/auth/complete?a')
        assert not client_config.check_redirect_url('http://my-site.com/auth/complete')

    def test_match_with_param(self):
        client_config = OpenIDClient(redirect_uris='http://my-site.com/auth/complete?a')
        assert client_config.check_redirect_url('http://my-site.com/auth/complete?a=2')
        assert client_config.check_redirect_url('http://my-site.com/auth/complete?a=3')
        assert client_config.check_redirect_url('http://my-site.com/auth/complete?a')
        assert not client_config.check_redirect_url('http://my-site.com/auth/complete')

    def test_match_without_defined_param(self):
        client_config = OpenIDClient(redirect_uris='http://my-site.com/auth/complete')
        assert not client_config.check_redirect_url('http://my-site.com/auth/complete?a=3')
        assert not client_config.check_redirect_url('http://my-site.com/auth/complete?a')




import sys



if __name__ == '__main__':

    import os
    os.environ['TEST_SERVER_CLIENT'] = 'yes'
    os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_settings'

    if os.path.exists('db.sqlite3'):
        os.remove('db.sqlite3')

    import django
    django.setup()

    from django.core.management import call_command
    from django.core.management import execute_from_command_line

    call_command('migrate')

    # create a sample user
    from django.contrib.auth.models import User
    u = User.objects.update_or_create(username='test',
                                      is_superuser=True,
                                      is_staff=True,
                                      defaults=dict(first_name='Johnny',
                                                    last_name='English',
                                                    email='johnny@english.co.uk'))[0]
    u.set_password('test')
    u.save()

    # create an OpenIDClient
    # python-social-auth passes authentication data in POST data
    from openid_connect_op.models import OpenIDClient
    from django.conf import settings
    client = OpenIDClient.objects.update_or_create(
        client_id=settings.KEY,
        defaults = dict(redirect_uris=settings.SERVER_URL + '/complete/test/',
                        client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_POST),
        client_name='Sample OpenID RP'
    )[0]

    client.set_client_secret(settings.SECRET)
    client.save()

    print('Will use redirect uri "%s"' % client.redirect_uris)

    args = ['manage.py', 'runserver']
    args.extend(sys.argv[1:])
    execute_from_command_line(args)



import json
from urllib.parse import splitquery

from django.conf import settings
from django.contrib.auth import logout
from django.http import HttpResponseRedirect
from django.http.response import HttpResponseBadRequest, HttpResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.http import urlencode
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from openid_connect_op.models import OpenIDClient, OpenIDKey, OpenIDToken
from openid_connect_op.signals import before_user_consent
from openid_connect_op.views.parameters import AuthenticationParameters
from . import OAuthRequestMixin
from .errors import OAuthError


class AuthenticationRequestView(OAuthRequestMixin, View):
    # noinspection PyAttributeOutsideInit
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        if request.method not in ('GET', 'POST'):
            return HttpResponseBadRequest('Only GET or POST are supported on OpenID endpoint')

        try:
            resp = self.unpack_auth_parameters(request)
            if resp:
                return resp

            self.validate_max_age()

            client = self._get_client_or_raise_exception()

            if self.should_login(request):
                return self.authenticate(request)

            should_sign_consent = self.should_request_user_consent(request, client)

            signal_responses = before_user_consent.send(type(client),
                                                        openid_client=client,
                                                        user=request.user,
                                                        consent_already_signed=not should_sign_consent)
            for resp in signal_responses:
                if isinstance(resp[1], HttpResponse):
                    return resp[1]

            if should_sign_consent:
                return self.request_user_consent(request, client)

            if 'code' in self.request_parameters.response_type:
                self.request_parameters.username = request.user.username
                auth_token, auth_db_token = \
                    OpenIDToken.create_token(client=client, token_type=OpenIDToken.TOKEN_TYPE_AUTH,
                                             token_data=self.request_parameters.to_dict(),
                                             ttl=getattr(settings, 'OPENID_AUTH_TOKEN_TTL', 20),
                                             user=request.user)
                return self.oauth_send_answer(request, {
                    'code': auth_token
                })
            else:
                raise OAuthError(error='parameter_not_supported',
                                 error_description='Only "code" value of the "response_type" '
                                                   'is supported by this server')

        except OAuthError as err:
            return self.oauth_send_answer(request, {
                'error': err.error,
                'error_description': err.error_description
            })

    def should_request_user_consent(self, request, client):
        return not client.has_user_agreement(request.user) or 'consent' in self.request_parameters.prompt

    def should_login(self, request):
        if request.user.is_anonymous or 'login' in self.request_parameters.prompt:
            if 'none' in self.request_parameters.prompt:
                raise OAuthError(error='login_required',
                                 error_description='No prompt requested but user is not logged in')
            return True
        return False

    def _get_client_or_raise_exception(self):
        try:
            client = OpenIDClient.objects.get(client_id=self.request_parameters.client_id)
        except OpenIDClient.DoesNotExist:
            raise OAuthError(error='unauthorized_client',
                             error_description='The client is unauthorized to get authentication token')
        if not client.check_redirect_url(self.request_parameters.redirect_uri):
            raise OAuthError(error='unauthorized_client',
                             error_description='The client is unauthorized to get authentication token '
                                               'as the redirect_uri does not match')
        return client

    def validate_max_age(self):
        if self.request_parameters.max_age:
            raise OAuthError(error='parameter_not_supported',
                             error_description='Parameter max_age is currently not supported by the server')

    def unpack_auth_parameters(self, request):
        try:
            if 'authp' in request.GET:
                self.request_parameters = AuthenticationParameters.unpack(
                    request.GET['authp'].encode('ASCII'), key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY))
            else:
                self.parse_request_parameters(request, AuthenticationParameters)
                if not hasattr(self.request_parameters, 'redirect_uri') or not self.request_parameters.redirect_uri:
                    return HttpResponseBadRequest(json.dumps({
                        'error': 'invalid_request',
                        'error_description': 'No redirect_uri in the request'
                    }))

                if not hasattr(self.request_parameters, 'client_id') or not self.request_parameters.client_id:
                    return HttpResponseBadRequest(json.dumps({
                        'error': 'invalid_request',
                        'error_description': 'No client_id in the request'
                    }))

                self.request_parameters.check_errors()
        except AttributeError as e:
            raise OAuthError(error='invalid_request', error_description=str(e))

    def authenticate(self, request):
        # if user is already logged in, a re-logging was requested, so forget the user
        if not request.user.is_anonymous:
            logout(request)

        this_url_with_encrypted_params = self.get_this_url_with_params(request)
        return HttpResponseRedirect(
            settings.LOGIN_URL + '?' + urlencode(
                {
                    'next': this_url_with_encrypted_params
                }
            )
        )

    def request_user_consent(self, request, client):
        this_url_with_encrypted_params = self.get_this_url_with_params(request)
        return HttpResponseRedirect(
            reverse(settings.OPENID_USER_CONSENT_VIEW, kwargs=dict(client_id=client.id)) + '?' + urlencode(
                {
                    'next': this_url_with_encrypted_params
                }
            )
        )

    def get_this_url_with_params(self, request):
        server, query = splitquery(request.build_absolute_uri())
        params = {
            'authp': self.request_parameters.pack(key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY))
        }
        return server + '?' + urlencode(params)



from django.http.response import JsonResponse, HttpResponse
from django.views import View

from openid_connect_op.utils.jwt import JWTTools


class JWKSView(View):
    def get(self, request, *args, **kwargs):
        return HttpResponse('{"keys": [%s]}' % JWTTools.get_jwks(), content_type='application/json')


import json

import pytest
from django.core.management import call_command


@pytest.mark.django_db
class TestWellKnownURL:

    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    def test_get_wellknown(self, client):
        resp = client.get('/.well-known/openid-configuration')
        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        print(data)
        assert data == {
            'authorization_endpoint': 'http://testserver/openid/authorize',
            'claim_types_supported': ['normal'],
            'claims_parameter_supported': False,
            'claims_supported': [
                'email', 'family_name', 'given_name', 'name', 'preferred_username',
                'sub', 'iss', 'auth_time', 'acr'
            ],
            'display_values_supported': ['page'],
            'id_token_signing_alg_values_supported': ['RS256'],
            'issuer': 'http://testserver/',
            'jwks_uri': 'http://testserver/openid/jwks',
            'response_types_supported': ['code'],
            'scopes_supported': ['openid', 'email', 'profile'],
            'subject_types_supported': ['public', 'pairwise'],
            'token_endpoint': 'http://testserver/openid/token',
            'end_session_endpoint': 'http://testserver/openid/logout',
            'registration_endpoint': 'http://testserver/openid/register',
            'token_endpoint_auth_methods_supported': ['client_secret_basic'],
            'ui_locales_supported': ['en-US', 'cs-CZ'],
            'userinfo_signing_alg_values_supported': ['RS256'],
            'userinfo_endpoint': 'http://testserver/openid/userinfo'
        }



import time

import pytest

from openid_connect_op.utils.crypto import CryptoTools

key = b'1234567890ABCDEF'


def test_crypt_decrypt():
    plaintext = b'Lorem ipsum dolor sit amet'
    crypto_text = CryptoTools.encrypt(plaintext, key=key)
    decrypted = CryptoTools.decrypt(crypto_text, key=key)
    assert decrypted == plaintext


def test_crypt_decrypt_timing():
    plaintext = b'Lorem ipsum dolor sit amet'
    crypto_text = CryptoTools.encrypt(plaintext, key=key, ttl=1)
    decrypted = CryptoTools.decrypt(crypto_text, key=key)
    assert decrypted == plaintext

    time.sleep(1)
    with pytest.raises(AttributeError) as e:
        CryptoTools.decrypt(crypto_text, key=key)
    assert 'The message has expired' == str(e.value)


def test_prefixes():
    plaintext = b'Lorem ipsum dolor sit amet'
    crypto_text = CryptoTools.encrypt(plaintext, key=key, prefix=b'abc')
    decrypted = CryptoTools.decrypt(crypto_text, key=key, expected_prefix=b'abc')
    assert decrypted == plaintext

    with pytest.raises(AttributeError) as e:
        CryptoTools.decrypt(crypto_text, key=key, expected_prefix=b'def')
    assert 'Expecting prefix b\'def\', got b\'abc\'' == str(e.value)



from django import forms
from django.http import HttpResponse
from django.utils import timezone
from django.utils.functional import cached_property
from django.views.generic import FormView

from openid_connect_op.models import OpenIDAgreement, OpenIDClient, OpenIDUserAgreement
from openid_connect_op.signals import after_user_consent


class ConsentForm(forms.Form):
    def __init__(self, *args, user=None, client=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = user
        self.agreements = list(client.get_user_agreements(user))
        self.agreed_agreements = set(OpenIDAgreement.objects.filter(client=client, user_agreements__user=user))
        for agreement in self.agreements:
            self.fields[self.agreement_field_name(agreement)] = \
                forms.BooleanField(required=False,
                                   label=agreement.text)
            agreement.agreed = agreement in self.agreed_agreements

    def agreement_field_name(self, agreement):
        return 'agreement_%s' % agreement.id

    def save(self):
        if not self.is_valid():
            raise ValueError('Can not save invalid data')
        for agreement in self.agreements:
            field_name = self.agreement_field_name(agreement)
            if agreement.obligatory or field_name in self.cleaned_data and self.cleaned_data[field_name]:
                # add the agreement if it does not exist
                OpenIDUserAgreement.objects.get_or_create(agreement=agreement, user=self.user, defaults={
                    'agreed_on': timezone.now(),
                    'agreed_by_user': True
                })
            else:
                # remove the agreement
                OpenIDUserAgreement.objects.filter(agreement=agreement, user=self.user).delete()


class ConsentView(FormView):
    template_name = 'django-open-id/consent.html'
    form_class = ConsentForm

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.request.user
        kwargs['client'] = self.openid_client
        return kwargs

    @cached_property
    def openid_client(self):
        return OpenIDClient.objects.get(pk=self.kwargs['client_id'])

    def get_context_data(self, **kwargs):
        return super().get_context_data(client=self.openid_client, **kwargs)

    def form_valid(self, form):
        form.save()
        signal_responses = after_user_consent.send(type(self.openid_client),
                                openid_client=self.openid_client,
                                user=self.request.user)
        for resp in signal_responses:
            if isinstance(resp[1], HttpResponse):
                return resp[1]
        return super().form_valid(form)

    def get_success_url(self):
        return self.request.GET.get('next') or '/'



try:
    import secrets
except ImportError:
    import openid_connect_op.utils.secrets_backport as secrets

import jwcrypto.jwk as jwk
from django.core.management import BaseCommand

from openid_connect_op.models import OpenIDClient, OpenIDKey


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('--redirect-url',
                            help='A set of urls separated by whitespace '
                                 'to which the authorization server can redirect to after authorization')
        parser.add_argument('--server-name',
                            help='Human name of the server')
        parser.add_argument('--auth-type',
                            help='Authentication type. Currently only basic or post are supported',
                            default='basic')
        parser.add_argument('--list',
                            help='List all registered clients', action="store_true")


    def handle(self, *args, **kwargs):

        if kwargs['list']:
            self.list()
            return

        if not kwargs['auth_type'] or not kwargs['server_name'] or not kwargs['redirect_url']:
            print("Required arguments: auth-type, server-name, redirect-url")
            return

        if kwargs['auth_type'] == 'basic':
            client_auth_type = OpenIDClient.CLIENT_AUTH_TYPE_BASIC
        elif kwargs['auth_type'] == 'post':
            client_auth_type = OpenIDClient.CLIENT_AUTH_TYPE_POST
        else:
            raise Exception('Only "post" or "basic" are supported for auth-type parameter')

        client_id     = secrets.token_urlsafe(64)
        client_secret = secrets.token_urlsafe(64)

        client = OpenIDClient.objects.get_or_create(client_id=client_id,
                                                    defaults={
                                                        'client_auth_type': client_auth_type,
                                                        'client_name': kwargs['server_name'],
                                                        'redirect_uris': kwargs['redirect_url']
                                                    })[0]
        client.set_client_secret(client_secret)
        client.save()

        print("""
Registration successfull, please configure the server with:
     Client ID (KEY in settings.py for python-social-auth)        : %s
     Client Secret (SECRET in settings.py for python-social-auth) : %s
""" % (client_id, client_secret))

    def list(self):
        for client in OpenIDClient.objects.all():
            print("client_id        ", client.client_id)
            print("human name       ", client.client_name)
            print("redirect_uris    ", client.redirect_uris)
            print("client auth      ", client.client_auth_type)
            print()


import base64
import hashlib
import json
from urllib.parse import urlencode, splitquery, parse_qs

import pytest
import time
from django.contrib.auth.models import User
from django.core.management import call_command

from openid_connect_op.models import OpenIDClient, OpenIDToken
from openid_connect_op.utils.jwt import JWTTools

BASIC_AUTH = 'Basic ' + base64.b64encode('test:b'.encode('utf-8')).decode('ascii')


@pytest.mark.django_db
class TestTokenRequest:

    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    @pytest.fixture
    def user(self):
        return User.objects.create(username='a')

    @pytest.fixture()
    def client_config(self):
        redirect_uri = 'http://localhost:8000/complete/test/?state=1234'
        ret = OpenIDClient.objects.create(
            client_id='test',
            redirect_uris=redirect_uri,
            client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_BASIC,
        )
        ret.set_client_secret('b')
        ret.save()
        return ret

    @pytest.fixture()
    def client_config_pairwise(self):
        redirect_uri = 'http://localhost:8000/complete/test/?state=1234'
        ret = OpenIDClient.objects.create(
            client_id='test',
            redirect_uris=redirect_uri,
            client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_BASIC,
            sub_hash='aaa'
        )
        ret.set_client_secret('b')
        ret.save()
        return ret

    def test_logged_user(self, client, client_config, user, settings):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        self.check_token_response(settings, client_config, resp)

    def test_double_authorization_code_for_user(self, client, client_config, user, settings):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        self.check_token_response(settings, client_config, resp)

        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        assert resp.status_code == 403
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {'error': 'unauthorized_client',
                        'error_description': 'Authorization token not found'}

    def test_logged_user_post(self, client, client_config, user, settings):
        # set auth type to POST
        client_config.client_auth_type = client_config.CLIENT_AUTH_TYPE_POST
        client_config.save()
        code = self.get_authorization_code(client, client_config, user)
        resp = client.post('/openid/token', {
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
            'client_id': 'test',
            'client_secret': 'b'
        })

        self.check_token_response(settings, client_config, resp)

    def test_logged_user_pairwise(self, client, client_config_pairwise, user, settings):
        code = self.get_authorization_code(client, client_config_pairwise, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config_pairwise.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        self.check_token_response(
            settings, client_config_pairwise, resp,
            sub=hashlib.sha256((user.username + client_config_pairwise.sub_hash).encode('utf-8')).hexdigest())

    @staticmethod
    def check_token_response(settings, client_config, resp, sub='a'):
        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert 'access_token' in data
        assert data['token_type'] == 'Bearer'
        assert 'refresh_token' in data
        assert data['expires_in'] == settings.OPENID_DEFAULT_ACCESS_TOKEN_TTL
        assert 'id_token' in data
        database_at = OpenIDToken.objects.get(token_hash=OpenIDToken.get_token_hash(data['access_token']))
        assert database_at.user.username == 'a'
        assert database_at.client == client_config
        assert database_at.token_type == OpenIDToken.TOKEN_TYPE_ACCESS_BEARER_TOKEN
        database_rt = OpenIDToken.objects.get(token_hash=OpenIDToken.get_token_hash(data['refresh_token']))
        assert database_rt.user.username == 'a'
        assert database_rt.client == client_config
        assert database_rt.token_type == OpenIDToken.TOKEN_TYPE_REFRESH_TOKEN
        # validate id token
        header, payload = JWTTools.validate_jwt(data['id_token'])
        assert header['alg'] == 'RS256'
        assert header['typ'] == 'JWT'
        assert payload['exp'] == int(payload['exp'])
        assert payload['iat'] == int(payload['iat'])
        assert payload['aud'] == ['test']
        assert payload['sub'] == sub  # username
        assert payload['iss'] == 'http://testserver/'

    @staticmethod
    def get_authorization_code(client, client_config, user):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code'
        }))
        assert resp.status_code == 302
        redirect_server, redirect_query = splitquery(resp.url)
        assert redirect_server == 'http://localhost:8000/complete/test/'
        redirect_query = parse_qs(redirect_query)
        assert redirect_query['state'] == ['1234']
        assert 'code' in redirect_query
        code = redirect_query['code'][0]
        return code

    def test_bad_redirect_uri(self, client, client_config, user):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_url': 'http://blah',
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)
        assert resp.status_code == 400
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {'error': 'invalid_request',
                        'error_description': 'redirect_uri does not match the one used in /authorize endpoint'}

    def test_no_grant_type(self, client, client_config, user):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_url': 'http://blah',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)
        assert resp.status_code == 400
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {'error': 'invalid_request',
                        'error_description': 'Required parameter with name "grant_type" is not present'}

    def test_bad_grant_type(self, client, client_config, user):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_url': 'http://blah',
            'grant_type': 'bad',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)
        assert resp.status_code == 400
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {'error': 'invalid_request',
                        'error_description': 'Value "bad" is not allowed for parameter grant_type. '
                                             'Allowed values are "authorization_code", "refresh_token"'}

    def test_bad_code(self, client, client_config, user):
        self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_url': 'http://blah',
            'grant_type': 'authorization_code',
            'code': '1234',
        }), HTTP_AUTHORIZATION=BASIC_AUTH)
        assert resp.status_code == 403
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {'error': 'unauthorized_client',
                        'error_description': 'Authorization token not found'}

    def test_no_code(self, client, client_config, user):
        self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_url': 'http://blah',
            'grant_type': 'authorization_code',
        }),
                          HTTP_AUTHORIZATION=BASIC_AUTH)
        assert resp.status_code == 400
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {'error': 'invalid_request',
                        'error_description': 'Required parameter with name "code" is not present'}

    def test_ok_refresh_user(self, client, client_config, user, settings):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        data = json.loads(resp.content.decode('utf-8'))
        refresh_token = data['refresh_token']

        resp = client.get('/openid/token?' + urlencode({
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)
        self.check_token_response(settings, client_config, resp)

    def test_refresh_no_token(self, client, client_config, user):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        data = json.loads(resp.content.decode('utf-8'))
        assert data['refresh_token'] is not None

        resp = client.get('/openid/token?' + urlencode({
            'grant_type': 'refresh_token',
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        assert resp.status_code == 400
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'error': 'invalid_request',
            'error_description': 'Required parameter with name "refresh_token" is not present'
        }

    def test_refresh_expired_token(self, client, client_config, user, settings):
        settings.OPENID_DEFAULT_ACCESS_TOKEN_TTL = 2
        settings.OPENID_DEFAULT_REFRESH_TOKEN_TTL = 4

        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        data = json.loads(resp.content.decode('utf-8'))
        refresh_token = data['refresh_token']
        time.sleep(settings.OPENID_DEFAULT_REFRESH_TOKEN_TTL + 1)
        resp = client.get('/openid/token?' + urlencode({
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        assert resp.status_code == 400
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'error': 'invalid_grant',
            'error_description': 'Refresh token expired'
        }
